import { renderChanges, setState } from "../../src/base/internal.js";
import WrappedStandardElement from "../../src/base/WrappedStandardElement.js";
import { assert } from "../testHelpers.js";

const WrappedA = WrappedStandardElement.wrap("a");
customElements.define("wrapped-a", WrappedA);

const WrappedButton = WrappedStandardElement.wrap("button");
customElements.define("wrapped-button", WrappedButton);

const WrappedDiv = WrappedStandardElement.wrap("div");
customElements.define("wrapped-div", WrappedDiv);

const WrappedImg = WrappedStandardElement.wrap("img");
customElements.define("wrapped-img", WrappedImg);

const WrappedInput = WrappedStandardElement.wrap("input");
customElements.define("wrapped-input", WrappedInput);

describe("WrappedStandardElement", () => {
  let container;

  before(() => {
    container = document.getElementById("container");
  });

  afterEach(() => {
    container.innerHTML = "";
  });

  it("creates an instance of the wrapped element", () => {
    const fixture = new WrappedA();
    fixture[renderChanges]();
    assert(fixture.inner instanceof HTMLAnchorElement);
  });

  it("exposes getter/setters that proxy to the wrapped element", () => {
    const fixture = new WrappedA();
    fixture.href = "http://localhost/foo/bar.html";
    container.append(fixture);
    assert.propertyVal(fixture.inner, "href", "http://localhost/foo/bar.html");
    assert.equal(fixture.protocol, "http:");
    assert.equal(fixture.hostname, "localhost");
    assert.equal(fixture.pathname, "/foo/bar.html");
  });

  it("marshals attributes to properties on the inner element", () => {
    const fixture = new WrappedA();
    container.append(fixture);
    fixture.setAttribute("href", "http://example.com/");
    fixture[renderChanges]();
    assert.propertyVal(fixture.inner, "href", "http://example.com/");
  });

  it("re-raises events not automatically retargetted by Shadow DOM", (done) => {
    const fixture = new WrappedImg();
    container.append(fixture);
    fixture.addEventListener("load", () => {
      done();
    });
    // Load a 1x1 pixel image to trigger the load event.
    fixture.src =
      "data:image/gif;base64,R0lGODlhAQABAIAAAAUEBAAAACwAAAAAAQABAAACAkQBADs=";
  });

  it("does not raise events if inner element is disabled", () => {
    const fixture = new WrappedButton();
    container.append(fixture);
    let count = 0;
    fixture.addEventListener("click", () => {
      count++;
    });
    fixture.click();
    fixture.disabled = true;
    fixture[renderChanges]();
    fixture.click();
    assert.equal(count, 1);
  });

  it("chooses an appropriate :host display style based on the wrapped element", () => {
    const fixtureA = new WrappedA();
    container.append(fixtureA);
    const fixtureDiv = new WrappedDiv();
    container.append(fixtureDiv);
    assert.equal(getComputedStyle(fixtureA).display, "inline-block");
    assert.equal(getComputedStyle(fixtureDiv).display, "block");
  });

  it("delegates boolean attributes", async () => {
    const fixture = new WrappedButton();
    container.append(fixture);

    // Disable via property.
    fixture.disabled = true;
    fixture[renderChanges]();
    assert.propertyVal(fixture.inner, "disabled", true);

    // // Re-enable via property.
    fixture.disabled = false;
    fixture[renderChanges]();
    assert.propertyVal(fixture.inner, "disabled", false);

    // Disable via attribute.
    fixture.setAttribute("disabled", "");
    fixture[renderChanges]();
    assert.propertyVal(fixture.inner, "disabled", true);

    // Re-enable via attribute.
    fixture.removeAttribute("disabled");
    fixture[renderChanges]();
    assert.propertyVal(fixture.inner, "disabled", false);
  });

  it("delegates tabindex state to inner element", async () => {
    const fixture = new WrappedInput();
    container.append(fixture);
    // NB: tabIndex is not part of WrappedInput's regular state; we're just
    // defining it. WrappedStandardElement should respect that.
    await fixture[setState]({ tabIndex: 1 });
    assert.equal(fixture.inner.tabIndex, 1);
  });

  it("delegates methods", async () => {
    const fixture = new WrappedInput();
    fixture.value = "Hello";
    container.append(fixture);
    fixture.setSelectionRange(1, 4);
    assert.equal(fixture.selectionStart, 1);
    assert.equal(fixture.selectionEnd, 4);
    fixture.focus();
    const root = fixture.shadowRoot;
    const activeElement = root && root.activeElement;
    assert.equal(activeElement, fixture.inner);
  });

  it("delegates attributes that don't correspond to properties", async () => {
    const fixture = new WrappedInput();
    fixture.setAttribute("readonly", "");
    fixture[renderChanges]();
    assert.equal(fixture.inner.getAttribute("readonly"), "");
  });
});
